local VERSION = 1.01

local AFFLICTION_SPEC_ID = 265

-- Patterns for parsing user input
local CMD_PATTERN = "^%s*(%a*)"
local NUMBER_PATTERN = "^%s*(%d*%.?%d+)"
local END_PATTERN = "^%s*$"

-- List of commands used for matching and help table
local CMD_UNLOCK = "unlock"
local CMD_LOCK = "lock"
local CMD_SIZE = "size"
local CMD_WIDTH = "width"
local CMD_HEIGHT = "height"
local CMD_SPACING = "spacing"
local CMD_INSET = "border"
local CMD_WARNING = "togglewarning"
local CMD_RESET = "reset"
local CMD_TOGGLE = "toggle"
local CMD_WARNING_COLOR = "warningcolor"
local CMD_FULL_COLOR = "fullcolor"
local CMD_BAR_COLOR = "barcolor"
local CMD_BACKGROUND_COLOR = "bgcolor"
local CMD_HELP = "help"

-- Errors for userinput
local INVALID_CMD_ERROR = "Invalid command."
local ARG_NUM_ERROR = "Invalid number of arguments for used command."
local ARG_TYPE_ERROR = "Input could not be parsed."
local ZERO_ERROR = "Zero is no valid input."

-- Table for the help command
local helpTable = {
  [CMD_UNLOCK] = "Unlocks the shards frame for positioning.",
  [CMD_LOCK] = "Locks the shards frame after unlocking with /bss ".. CMD_UNLOCK .. ".",
  [CMD_SIZE] = "Lets you adjust the size of the shards by entering two numbers, first the width, second the height, e.g. /bss ".. CMD_SIZE .. " 30 45",
  [CMD_WIDTH] = "Lets you adjust the width of the shards.",
  [CMD_HEIGHT] = "Lets you adjust the height of the shards.",
  [CMD_SPACING] = "Lets you adjust the distance between the shards.",
  [CMD_INSET] = "Lets you adjust the thickness of the border.",
  [CMD_WARNING] = "Toggles the cap warning.",
  [CMD_RESET] = "Resets the size, position and colors of the shards.",
  [CMD_TOGGLE] = "Toggles the display of the shards.",
  [CMD_WARNING_COLOR] = "Lets you pick the color for the warning.",
  [CMD_FULL_COLOR] = "Lets you adjust the color for full shards.",
  [CMD_BAR_COLOR] = "Lets you adjust the color of filling shards.",
  [CMD_BACKGROUND_COLOR] = "Lets you adjust the background color.",
  [CMD_HELP] = "Prints the help function.",
}

local WELCOME_MSG = "Welcome to BetterSoulShards! For information on how to configurate this addon type: /bss help"

-- If the number of shards should change, a talent gets implemented, use this value.
local NUM_SHARDS = 4

local inputError = false
local combat = false
local spec = nil
local db = nil


-- Initializing some tables
bss_shard = {{},{},{},{}}   
local bssOnUpdate = {}

-- Setting up the slashcommand
SLASH_BSS1 = "/bss"
SlashCmdList["BSS"] = 
  function(msg)
    inputError = false
    slashHandler(bssParseInput(string.lower(msg)))
  end

-- Parser function for slash commands
-- Arguments for the slash command must be of the form /bss string number number number ...
function bssParseInput(msg) 
  local _, idx, cmd = msg:find(CMD_PATTERN)
  if not cmd then
    print(INVALID_CMD_ERROR)
    inputError = true
    return
  else
    return cmd, bssParseNumbers(msg, idx + 1)
  end
end

-- Parses the numbers after the initial command argument
function bssParseNumbers(msg, idx)
  if (msg:match(END_PATTERN, idx) == "") then
    return
  else
    local _, fin, number = msg:find(NUMBER_PATTERN, idx)
    if not number then
      print(ARG_TYPE_ERROR)
      inputError = true
      return
    else
      return number, bssParseNumbers(msg, fin + 1)
    end  
  end
end

-- Reset the mainframe's position and mouse interactivity.
local function resetMF()
  bssMainFrame:ClearAllPoints()
  bssMainFrame:EnableMouse(false)
  bssMainFrame:SetPoint("CENTER", UIParent ,"CENTER")
  bssMainFrame:SetUserPlaced(false)
end

-- Reset the SavedVariablesPerCharacter
local function resetDB()

  db = BetterSoulShardsDB or {}
  
  db.newChar = true
  
  db.version = VERSION
  
  -- Initializing shape
  db.width = 30
  db.height = 45
  db.spacing = 10
  db.inset = 4
  
  db.active = true
  db.capWarning = true
  
  -- Initializing colors
  db.color = { 
    progressBar = {
      red = 0.55,
      green = 0.0,
      blue = 1.0,
      alpha = 0.5
    },
    
    bg = {
      red = 0.0,
      green = 0.0,
      blue = 0.0,
      alpha = 1.0      
    },
    
    full = {
      red = 0.55,
      green = 0.0,
      blue = 1.0,
      alpha = 1.0
    },
    
    capWarning = {
      red = 0.5,
      green = 0.0,
      blue = 0.0,
      alpha = 1.0
    }
     
  }
end

local function updateDB()
  if not db.version then
    resetDB()
  end
end

local function bssOnEvent(self, event, ...)
  if event == "ADDON_LOADED" and ... == "BetterSoulShards" then
    if not BetterSoulShardsDB then
      resetDB()
      BetterSoulShardsDB = db
    end
    db = BetterSoulShardsDB
    self:UnregisterEvent(event)
    bssInitialize()
    if not spec then 
      bssMainFrame:Hide()
    end
  elseif event == "PLAYER_REGEN_DISABLED" then
    combat = true
  elseif event == "PLAYER_REGEN_ENABLED" then
    combat = false
  elseif event == "ACTIVE_TALENT_GROUP_CHANGED" then
    spec = GetSpecialization()
    if (db.active and spec and (GetSpecializationInfo(spec) == AFFLICTION_SPEC_ID)) then
      bssMainFrame:Show()
      print(WELCOME_MSG)
    else
      bssMainFrame:Hide()
    end
  elseif event == "PLAYER_ENTERING_WORLD" then
    spec = GetSpecialization()
    print(WELCOME_MSG)
    if (db.active and spec and (GetSpecializationInfo(spec) == AFFLICTION_SPEC_ID)) then
      bssMainFrame:Show()
      if newChar then
      end
    else
      bssMainFrame:Hide()
    end
  end
end


-- Run once upon "ADDON_LOADED" event
function bssInitialize()
  -- Define function to update background color depending on capWarning and current powerlevel
  local updateBGColor = function (self, power) 
    if db.capWarning then
        if power > 275 and combat then
          self.bg:SetTexture(db.color.capWarning.red, db.color.capWarning.green, db.color.capWarning.blue, db.color.capWarning.alpha)
        else
          self.bg:SetTexture(db.color.bg.red, db.color.bg.green, db.color.bg.blue, db.color.bg.alpha)
        end
      end
  end
  
  -- Define function to update progressbar color
  local updateBarColor = function (self) 
    self.progressBar:SetStatusBarTexture(db.color.progressBar.red, db.color.progressBar.green, db.color.progressBar.blue, db.color.progressBar.alpha)
  end
  
  -- Define function to update color for full shards
  local updateFullColor = function (self) 
    self.full:SetTexture(db.color.full.red, db.color.full.green, db.color.full.blue, db.color.full.alpha)
  end
  
  -- If version is newer than the database, handle the differences
  if not db.version or db.version < VERSION then
    updateDB()
  end
  
  for i = 1,NUM_SHARDS do
    -- Setting up each shard
    bss_shard[i].progressBar = CreateFrame("StatusBar", "bssProgressBar" .. i, bssMainFrame)
    bss_shard[i].bg = bss_shard[i].progressBar:CreateTexture("bssBG1", "BACKGROUND")
    bss_shard[i].full = bss_shard[i].progressBar:CreateTexture("bssFull1", "OVERLAY")
    bss_shard[i].progressBar:EnableMouse(false)
    
    -- Assign color update functions
    bss_shard[i].updateBGColor = updateBGColor
    bss_shard[i].updateBarColor = updateBarColor
    bss_shard[i].updateFullColor = updateFullColor
    
    bssOnUpdate[i] = function (self, ...)
      local power = UnitPower("player", 7, true)
      if (power >= i*100) then
        bss_shard[i].full:Show()
      else
        bss_shard[i].full:Hide()
      end
      bss_shard[i].progressBar:SetValue(power)
      bss_shard[i]:updateBGColor(power)
    end
    
    bss_shard[i].progressBar:SetScript("OnUpdate", bssOnUpdate[i])
  end
  bssUpdate()
end

-- Update the frames with date from the database
function bssUpdate()
  bssMainFrame:SetSize(db.width * NUM_SHARDS + db.spacing * (NUM_SHARDS - 1) + 30, db.height + 30)
  if db.active then
    bssMainFrame:Show()
  end
  for i = 1,NUM_SHARDS do
    
    bss_shard[i].progressBar:SetSize(db.width-db.inset, db.height-db.inset)
    bss_shard[i].progressBar:SetPoint("LEFT", bssMainFrame ,"LEFT", 15 + (i - 1)*(db.width + db.spacing), 0)
    bss_shard[i].progressBar:SetMinMaxValues((i - 1)* 100, i * 100)
    
    bss_shard[i].bg:SetSize(db.width, db.height)
    bss_shard[i].bg:SetPoint("CENTER", bss_shard[i].progressBar, "CENTER")
    
    bss_shard[i].full:SetSize(db.width-db.inset, db.height-db.inset)
    bss_shard[i].full:SetPoint("CENTER", bss_shard[i].progressBar)
    
    bss_shard[i]:updateBarColor()
    bss_shard[i]:updateFullColor()
  end
end

-- Opens up the color picker frame for the layer
local function showColorPicker(layer)
  local callBackFunction = bssGenerateColorCallBack(layer)
  local a, r, g, b = db.color[layer].alpha, db.color[layer].red, db.color[layer].green, db.color[layer].blue
  ColorPickerFrame.func, ColorPickerFrame.opacityFunc, ColorPickerFrame.cancelFunc = callBackFunction, callBackFunction, callBackFunction
  ColorPickerFrame:SetColorRGB(r,g,b)
  ColorPickerFrame.opacity = (1 - a)
  ColorPickerFrame.hasOpacity = true
  ColorPickerFrame.previousValues = {a, r, g, b}
  ColorPickerFrame:Show()
end

-- Generates a callback function that handles updating of the shards upon user input to the color picker
function bssGenerateColorCallBack(layer)
  local function res(restore)
    if restore then
      bssSetDBColor(layer, unpack(restore))
    else
      bssSetDBColor(layer, (1 - OpacitySliderFrame:GetValue()), ColorPickerFrame:GetColorRGB())
    end
    bssUpdate()
  end
  return res
end

function bssSetDBColor(layer, a, r, g, b)
  db.color[layer].red = r
  db.color[layer].green = g
  db.color[layer].blue = b
  db.color[layer].alpha = a or db.color[layer].alpha
end

function slashHandler(cmd, ...)
  if (cmd == CMD_UNLOCK) then
    bssMainFrame:EnableMouse(true)
  elseif (cmd == CMD_LOCK) then
    bssMainFrame:EnableMouse(false)
    -- bssMainFrame:SetUserPlaced(false)
  elseif (cmd == CMD_SIZE) and (select("#", ...) == 2) then
    local width, height = ...
    if (width * height) == 0 then
      print(ZERO_ERROR)
    else
      db.width, db.height = width, height
    end
  elseif (cmd == CMD_WIDTH) and (select("#", ...) == 1) then
    local width = ...
    if width * 1 == 0 then
      print(ZERO_ERROR)
    else
      db.width = width
    end
  elseif (cmd == CMD_HEIGHT) and (select("#", ...) == 1) then
    local height = ...
    if height * 1 == 0 then
      print(ZERO_ERROR)
    else
      db.height = height
    end
  elseif (cmd == CMD_SPACING) and (select("#", ...) == 1) then
    db.spacing = ...
  elseif (cmd == CMD_INSET) and (select("#", ...) == 1) then
    db.inset = ...
  elseif (cmd == CMD_WARNING) then
    db.capWarning = not db.capWarning
  elseif (cmd == CMD_RESET) then
    resetMF()
    resetDB()
    print(WELCOME_MSG)
  elseif (cmd == CMD_TOGGLE) then
    if db.active then
      bssMainFrame:Hide()
    else
      bssMainFrame:Show()
    end
    db.active = not db.active
  elseif (cmd == CMD_WARNING_COLOR) then
    local r, g, b, a = ...
    if r and g and b then
      bssSetDBColor("capWarning", a, r, g, b)
    else
      showColorPicker("capWarning")
    end
  elseif (cmd == CMD_FULL_COLOR) then
    local r, g, b, a = ...
    if r and g and b then
      bssSetDBColor("full", a, r, g, b)
    else
      showColorPicker("full")
    end
  elseif (cmd == CMD_BAR_COLOR) then
    local r, g, b, a = ...
    if r and g and b then
      bssSetDBColor("progressBar", a, r, g, b)
    else
      showColorPicker("progressBar")
    end
  elseif (cmd == CMD_BACKGROUND_COLOR) then
    local r, g, b, a = ...
    if r and g and b then
      bssSetDBColor("bg", a, r, g, b)
    else
      showColorPicker("bg")
    end
  elseif (cmd == CMD_HELP) then
    local format = string.format
    print("For coloradjustments either add r,g,b[,alpha] (within the range of 0.0 to 1.0) as arguments to the command or leave arguments blank to open the color picker.")
    for cmd, help in pairs(helpTable) do
      print(format("/bss %s : %s",cmd, help))
    end
  else
    print(INVALID_CMD_ERROR)
  end
  db.newChar = false
  bssUpdate()
end

CreateFrame("BUTTON", "bssMainFrame", UIParent)
bssMainFrame:Hide()
bssMainFrame:SetMovable(true)
bssMainFrame:EnableMouse(false)

bssMainFrame:SetPoint("CENTER", UIParent ,"CENTER")

bssMainFrame:SetScript("OnEvent", bssOnEvent)
bssMainFrame:RegisterEvent("ADDON_LOADED")
bssMainFrame:RegisterEvent("PLAYER_REGEN_DISABLED")
bssMainFrame:RegisterEvent("PLAYER_REGEN_ENABLED")
bssMainFrame:RegisterEvent("ACTIVE_TALENT_GROUP_CHANGED")
bssMainFrame:RegisterEvent("PLAYER_ENTERING_WORLD")
bssMainFrame:RegisterForDrag("LeftButton")
bssMainFrame:SetScript("OnDragStart", function(self, button) self:StartMoving() end)
bssMainFrame:SetScript("OnDragStop", function(self) self:StopMovingOrSizing() end)